"""CIB XML generator for Pacemaker's Cluster Test Suite (CTS)."""

__all__ = [
    "Alerts",
    "Clone",
    "Expression",
    "FencingTopology",
    "Group",
    "Nodes",
    "OpDefaults",
    "Option",
    "Resource",
    "Rule",
]
__copyright__ = "Copyright 2008-2025 the Pacemaker project contributors"
__license__ = "GNU General Public License version 2 or later (GPLv2+) WITHOUT ANY WARRANTY"


def key_val_string(**kwargs):
    """
    Construct a string from kwargs containing key=value pairs separated by spaces.

    This is suitable for using in an XML element as a list of its attributes.
    Any pairs that have value=None will be skipped.

    Note that a dictionary can be passed to this function instead of kwargs
    by using a construction like:

        key_val_string(**{"a": 1, "b": 2})
    """
    retval = ""

    for (k, v) in kwargs.items():
        if v is None:
            continue

        retval += f' {k}="{v}"'

    return retval


def element(element_name, **kwargs):
    """
    Create an XML element string with the given element_name and attributes.

    This element does not support having any children, so it will be closed
    on the same line.  The attributes are processed by key_val_string.
    """
    return f"<{element_name} {key_val_string(**kwargs)}/>"


def containing_element(element_name, inner, **kwargs):
    """Like element, but surrounds some child text passed by the inner parameter."""
    attrs = key_val_string(**kwargs)
    return f"<{element_name} {attrs}>{inner}</{element_name}>"


class XmlBase:
    """
    A base class for deriving all kinds of XML sections in the CIB.

    This class contains only the most basic operations common to all sections.
    It is up to subclasses to provide most behavior.

    Note that subclasses of this base class often have different sets of
    arguments to their __init__ methods.  In general this is not a great
    practice, however it is so thoroughly used in these classes that trying
    to straighten it out is likely to cause more bugs than just leaving it
    alone for now.
    """

    def __init__(self, factory, tag, _id, **kwargs):
        """
        Create a new XmlBase instance.

        Arguments:
        factory -- A ConfigFactory instance
        tag     -- The XML element's start and end tag
        _id     -- A unique name for the element
        kwargs  -- Any additional key/value pairs that should be added to
                   this element as attributes
        """
        self._children = []
        self._factory = factory
        self._kwargs = kwargs
        self._tag = tag

        self.name = _id

    def __repr__(self):
        """Return a short string description of this XML section."""
        return f"{self._tag}-{self.name}"

    def add_child(self, child):
        """Add an XML section as a child of this one."""
        self._children.append(child)

    def __setitem__(self, key, value):
        """
        Add a key/value pair to this element.

        The resulting pair becomes an XML attribute.  If value is None, remove
        the key.
        """
        if value:
            self._kwargs[key] = value
        else:
            self._kwargs.pop(key, None)

    def show(self):
        """Recursively return a string representation of this XML section."""
        text = f"<{self._tag}"
        if self.name:
            text += f' id="{self.name}"'

        text += key_val_string(**self._kwargs)

        if not self._children:
            text += '''/>'''
            return text

        text += '''>'''

        for c in self._children:
            text += c.show()

        text += f"</{self._tag}>"
        return text

    def _run(self, operation, xml, section, options=""):
        """
        Update the CIB on the cluster to include this XML section.

        Arguments:
        operation -- Whether this update is a "create" or "modify" operation
        xml       -- The XML to update the CIB with, typically the result
                     of calling show
        section   -- Which section of the CIB this update applies to (see
                     the --scope argument to cibadmin for allowed values)
        options   -- Extra options to pass to cibadmin
        """
        if self.name:
            label = self.name
        else:
            label = f"<{self._tag}>"

        self._factory.debug(f"Writing out {label}")

        fixed = f"HOME=/root CIB_file={self._factory.tmpfile}"
        fixed += f" cibadmin --{operation} --scope {section} {options} --xml-text '{xml}'"

        (rc, _) = self._factory.rsh(self._factory.target, fixed)
        if rc != 0:
            raise RuntimeError(f"Configure call failed: {fixed}")


class InstanceAttributes(XmlBase):
    """Create an <instance_attributes> XML section with key/value pairs."""

    def __init__(self, factory, _id, attrs):
        """
        Create a new InstanceAttributes instance.

        Arguments:
        factory -- A ConfigFactory instance
        _id     -- A unique name for the element
        attrs   -- Key/value pairs to add as nvpair child elements
        """
        XmlBase.__init__(self, factory, "instance_attributes", _id)

        # Create an <nvpair> for each attribute
        for (attr, value) in attrs.items():
            self.add_child(XmlBase(factory, "nvpair", f"{_id}-{attr}",
                                   name=attr, value=value))


class Node(XmlBase):
    """Create a <node> XML section for a single node, complete with node attributes."""

    def __init__(self, factory, node_name, node_id, node_attrs):
        """
        Create a new Node instance.

        Arguments:
        factory    -- A ConfigFactory instance
        node_name  -- The value of the uname attribute for this node
        node_id    -- A unique name for the element
        node_attrs -- Additional key/value pairs to set as instance
                      attributes for this node
        """
        XmlBase.__init__(self, factory, "node", node_id, uname=node_name)
        self.add_child(InstanceAttributes(factory, f"{node_name}-1", node_attrs))


class Nodes(XmlBase):
    """Create a <nodes> XML section containing multiple Node instances as children."""

    def __init__(self, factory):
        """
        Create a new Nodes instance.

        Arguments:
        factory -- A ConfigFactory instance
        """
        XmlBase.__init__(self, factory, "nodes", None)

    def add_node(self, node_name, node_id, node_attrs):
        """
        Add a child node element.

        Arguments:
        node_name  -- The value of the uname attribute for this node
        node_id    -- A unique name for the element
        node_attrs -- Additional key/value pairs to set as instance
                      attributes for this node
        """
        self.add_child(Node(self._factory, node_name, node_id, node_attrs))

    def commit(self):
        """Modify the CIB on the cluster to include this XML section."""
        self._run("modify", self.show(), "configuration", "--allow-create")


class FencingTopology(XmlBase):
    """Create a <fencing-topology> XML section describing how fencing is configured in the cluster."""

    def __init__(self, factory):
        """
        Create a new FencingTopology instance.

        Arguments:
        factory -- A ConfigFactory instance
        """
        XmlBase.__init__(self, factory, "fencing-topology", None)

    def level(self, index, target, devices, target_attr=None, target_value=None):
        """
        Generate a <fencing-level> XML element.

        index        -- The order in which to attempt fencing-levels
                        (1 through 9).  Levels are attempted in ascending
                        order until one succeeds.
        target       -- The name of a single node to which this level applies
        devices      -- A list of devices that must all be tried for this
                        level
        target_attr  -- The name of a node attribute that is set for nodes
                        to which this level applies
        target_value -- The value of a node attribute that is set for nodes
                        to which this level applies
        """
        if target:
            xml_id = f"cts-{target}.{index}"
            self.add_child(XmlBase(self._factory, "fencing-level", xml_id, target=target, index=index, devices=devices))

        else:
            xml_id = f"{target_attr}-{target_value}.{index}"
            child = XmlBase(self._factory, "fencing-level", xml_id, index=index, devices=devices)
            child["target-attribute"] = target_attr
            child["target-value"] = target_value
            self.add_child(child)

    def commit(self):
        """Create this XML section in the CIB."""
        self._run("create", self.show(), "configuration", "--allow-create")


class Option(XmlBase):
    """Create a <cluster_property_set> XML section of key/value pairs for cluster-wide configuration settings."""

    def __init__(self, factory, _id="cib-bootstrap-options"):
        """
        Create a new Option instance.

        Arguments:
        factory -- A ConfigFactory instance
        _id     -- A unique name for the element
        """
        XmlBase.__init__(self, factory, "cluster_property_set", _id)

    def __setitem__(self, key, value):
        """Add a child nvpair element containing the given key/value pair."""
        self.add_child(XmlBase(self._factory, "nvpair", f"cts-{key}", name=key, value=value))

    def commit(self):
        """Modify the CIB on the cluster to include this XML section."""
        self._run("modify", self.show(), "crm_config", "--allow-create")


class OpDefaults(XmlBase):
    """Create a <cts-op_defaults-meta> XML section of key/value pairs for operation default settings."""

    def __init__(self, factory):
        """
        Create a new OpDefaults instance.

        Arguments:
        factory -- A ConfigFactory instance
        """
        XmlBase.__init__(self, factory, "op_defaults", None)
        self.meta = XmlBase(self._factory, "meta_attributes", "cts-op_defaults-meta")
        self.add_child(self.meta)

    def __setitem__(self, key, value):
        """Add a child nvpair meta_attribute element containing the given key/value pair."""
        self.meta.add_child(XmlBase(self._factory, "nvpair", f"cts-op_defaults-{key}", name=key, value=value))

    def commit(self):
        """Modify the CIB on the cluster to include this XML section."""
        self._run("modify", self.show(), "configuration", "--allow-create")


class Alerts(XmlBase):
    """Create an <alerts> XML section."""

    def __init__(self, factory):
        """
        Create a new Alerts instance.

        Arguments:
        factory -- A ConfigFactory instance
        """
        XmlBase.__init__(self, factory, "alerts", None)
        self._alert_count = 0

    def add_alert(self, path, recipient):
        """
        Create a new alert as a child of this XML section.

        Arguments:
        path      -- The path to a script to be called when a cluster
                     event occurs
        recipient -- An environment variable to be passed to the script
        """
        self._alert_count += 1
        alert = XmlBase(self._factory, "alert", f"alert-{self._alert_count}",
                        path=path)
        recipient1 = XmlBase(self._factory, "recipient",
                             f"alert-{self._alert_count}-recipient-1",
                             value=recipient)
        alert.add_child(recipient1)
        self.add_child(alert)

    def commit(self):
        """Modify the CIB on the cluster to include this XML section."""
        self._run("modify", self.show(), "configuration", "--allow-create")


class Expression(XmlBase):
    """Create an <expression> XML element as part of some constraint rule."""

    def __init__(self, factory, _id, attr, op, value=None):
        """
        Create a new Expression instance.

        Arguments:
        factory -- A ConfigFactory instance
        _id     -- A unique name for the element
        attr    -- The attribute to be tested
        op      -- The comparison to perform ("lt", "eq", "defined", etc.)
        value   -- Value for comparison (can be None for "defined" and
                   "not_defined" operations)
        """
        XmlBase.__init__(self, factory, "expression", _id, attribute=attr, operation=op)
        if value:
            self["value"] = value


class Rule(XmlBase):
    """Create a <rule> XML section consisting of one or more expressions, as part of some constraint."""

    def __init__(self, factory, _id, score, op="and", expr=None):
        """
        Create a new Rule instance.

        Arguments:
        factory -- A ConfigFactory instance
        _id     -- A unique name for the element
        score   -- If this rule is used in a location constraint and
                   evaluates to true, apply this score to the constraint
        op      -- If this rule contains more than one expression, use this
                   boolean op when evaluating
        expr    -- An Expression instance that can be added to this Rule
                   when it is created
        """
        XmlBase.__init__(self, factory, "rule", _id)

        self["boolean-op"] = op
        self["score"] = score

        if expr:
            self.add_child(expr)


class Resource(XmlBase):
    """
    A base class that creates all kinds of <resource> XML sections.

    These sections fully describe a single cluster resource.  This defaults to
    primitive resources, but subclasses can create other types.
    """

    def __init__(self, factory, _id, rtype, standard, provider=None):
        """
        Create a new Resource instance.

        Arguments:
        factory  -- A ConfigFactory instance
        _id      -- A unique name for the element
        rtype    -- The name of the resource agent
        standard -- The standard the resource agent follows ("ocf",
                    "systemd", etc.)
        provider -- The vendor providing the resource agent
        """
        XmlBase.__init__(self, factory, "native", _id)

        self._provider = provider
        self._rtype = rtype
        self._standard = standard

        self._meta = {}
        self._op = []
        self._param = {}

        self._coloc = {}
        self._needs = {}
        self._scores = {}

        if self._standard == "ocf" and not provider:
            self._provider = "heartbeat"
        elif self._standard == "lsb":
            self._provider = None

    def __setitem__(self, key, value):
        """Add a child nvpair element containing the given key/value pair as an instance attribute."""
        self._add_param(key, value)

    def add_op(self, _id, interval, **kwargs):
        """
        Add an operation child XML element to this resource.

        Arguments:
        _id      -- A unique name for the element.  Also, the action to
                    perform ("monitor", "start", "stop", etc.)
        interval -- How frequently (in seconds) to perform the operation
        kwargs   -- Any additional key/value pairs that should be added to
                    this element as attributes
        """
        self._op.append(XmlBase(self._factory, "op", f"{_id}-{interval}",
                                name=_id, interval=interval, **kwargs))

    def _add_param(self, name, value):
        """Add a child nvpair element containing the given key/value pair as an instance attribute."""
        self._param[name] = value

    def add_meta(self, name, value):
        """Add a child nvpair element containing the given key/value pair as a meta attribute."""
        self._meta[name] = value

    def prefer(self, node, score="INFINITY", rule=None):
        """
        Add a location constraint where this resource prefers some node.

        Arguments:
        node  -- The name of the node to prefer
        score -- Apply this score to the location constraint
        rule  -- A Rule instance to use in creating this constraint, instead
                 of creating a new rule
        """
        if not rule:
            rule = Rule(self._factory, f"prefer-{node}-r", score,
                        expr=Expression(self._factory, f"prefer-{node}-e", "#uname", "eq", node))

        self._scores[node] = rule

    def after(self, resource, kind="Mandatory", first="start", then="start", **kwargs):
        """
        Create an ordering constraint between this resource and some other.

        Arguments:
        resource -- The name of the dependent resource
        kind     -- How to enforce the constraint ("mandatory", "optional",
                    "serialize")
        first    -- The action that this resource must complete before the
                    then-action can be initiated for the dependent resource
                    ("start", "stop", "promote", "demote")
        then     -- The action that the dependent resource can execute only
                    after the first-action has completed (same values as
                    first)
        kwargs   -- Any additional key/value pairs that should be added to
                    this element as attributes
        """
        kargs = kwargs.copy()
        kargs["kind"] = kind

        if then:
            kargs["first-action"] = "start"
            kargs["then-action"] = then

        if first:
            kargs["first-action"] = first

        self._needs[resource] = kargs

    def colocate(self, resource, score="INFINITY", role=None, withrole=None, **kwargs):
        """
        Create a colocation constraint between this resource and some other.

        Arguments:
        resource -- The name of the resource that should be located relative
                    this one
        score    -- Apply this score to the colocation constraint
        role     -- Apply this colocation constraint only to promotable clones
                    in this role ("started", "promoted", "unpromoted")
        withrole -- Apply this colocation constraint only to with-rsc promotable
                    clones in this role
        kwargs   -- Any additional key/value pairs that should be added to
                    this element as attributes
        """
        kargs = kwargs.copy()
        kargs["score"] = score

        if role:
            kargs["rsc-role"] = role

        if withrole:
            kargs["with-rsc-role"] = withrole

        self._coloc[resource] = kargs

    def _constraints(self):
        """Generate a <constraints> XML section containing all previously added ordering and colocation constraints."""
        text = "<constraints>"

        for (k, v) in self._scores.items():
            attrs = {"id": f"prefer-{k}", "rsc": self.name}
            text += containing_element("rsc_location", v.show(), **attrs)

        for (k, kargs) in self._needs.items():
            attrs = {"id": f"{self.name}-after-{k}", "first": k, "then": self.name}
            text += element("rsc_order", **attrs, **kargs)

        for (k, kargs) in self._coloc.items():
            attrs = {"id": f"{self.name}-with-{k}", "rsc": self.name, "with-rsc": k}
            text += element("rsc_colocation", **attrs)

        text += "</constraints>"
        return text

    def show(self):
        """Recursively return a string representation of this XML section."""
        text = f'<primitive id="{self.name}" class="{self._standard}" type="{self._rtype}"'

        if self._provider:
            text += f' provider="{self._provider}"'

        text += '''>'''

        if self._meta:
            nvpairs = ""
            for (p, v) in self._meta.items():
                attrs = {"id": f"{self.name}-{p}", "name": p, "value": v}
                nvpairs += element("nvpair", **attrs)

            text += containing_element("meta_attributes", nvpairs,
                                       id=f"{self.name}-meta")

        if self._param:
            nvpairs = ""
            for (p, v) in self._param.items():
                attrs = {"id": f"{self.name}-{p}", "name": p, "value": v}
                nvpairs += element("nvpair", **attrs)

            text += containing_element("instance_attributes", nvpairs,
                                       id=f"{self.name}-params")

        if self._op:
            text += '''<operations>'''

            for o in self._op:
                key = o.name
                o.name = f"{self.name}-{key}"
                text += o.show()
                o.name = key

            text += '''</operations>'''

        text += '''</primitive>'''
        return text

    def commit(self):
        """Modify the CIB on the cluster to include this XML section."""
        self._run("create", self.show(), "resources")
        self._run("modify", self._constraints(), "constraints")


class Group(Resource):
    """
    A specialized Resource subclass that creates a <group> XML section.

    This section describes a single group resource consisting of multiple child
    primitive resources.
    """

    def __init__(self, factory, _id):
        """
        Create a new Group instance.

        Arguments:
        factory -- A ConfigFactory instance
        _id     -- A unique name for the element
        """
        Resource.__init__(self, factory, _id, None, None)
        self.tag = "group"

    def __setitem__(self, key, value):
        """Add a child nvpair element containing the given key/value pair as an instance attribute."""
        self.add_meta(key, value)

    def show(self):
        """Recursively return a string representation of this XML section."""
        text = f'<{self.tag} id="{self.name}">'

        if len(self._meta) > 0:
            nvpairs = ""
            for (p, v) in self._meta.items():
                attrs = {"id": f"{self.name}-{p}", "name": p, "value": v}
                nvpairs += element("nvpair", **attrs)

            text += containing_element("meta_attributes", nvpairs,
                                       id=f"{self.name}-meta")

        for c in self._children:
            text += c.show()

        text += f"</{self.tag}>"
        return text


class Clone(Group):
    """
    A specialized Group subclass that creates a <clone> XML section.

    This section describes a clone resource containing multiple instances of a
    single primitive resource.
    """

    def __init__(self, factory, _id, child=None):
        """
        Create a new Clone instance.

        Arguments:
        factory -- A ConfigFactory instance
        _id     -- A unique name for the element
        child   -- A Resource instance that can be added to this Clone
                   when it is created.  Alternately, use add_child later.
                   Note that a Clone may only have one child.
        """
        Group.__init__(self, factory, _id)
        self.tag = "clone"

        if child:
            self.add_child(child)

    def add_child(self, child):
        """
        Add the given resource as a child of this Clone.

        Note that a Clone resource only supports one child at a time.
        """
        if not self._children:
            self._children.append(child)
        else:
            self._factory.log(f"Clones can only have a single child. Ignoring {child.name}")



"""Logging classes for Pacemaker's Cluster Test Suite (CTS)."""

__all__ = ["LogFactory"]
__copyright__ = "Copyright 2014-2025 the Pacemaker project contributors"
__license__ = "GNU General Public License version 2 or later (GPLv2+) WITHOUT ANY WARRANTY"

import os
import sys
import time


class Logger:
    """Abstract class to use as parent for CTS logging classes."""

    TimeFormat = "%b %d %H:%M:%S\t"

    def __init__(self, filename=None, tag=None):
        # Whether this logger should print debug messages
        self._debug_target = True

        self._logfile = filename

        if tag:
            self._source = f"{tag}: "
        else:
            self._source = ""

    def __call__(self, lines):
        """Log specified messages."""
        raise ValueError("Abstract class member (__call__)")

    def write(self, line):
        """Log a single line excluding trailing whitespace."""
        return self(line.rstrip())

    def writelines(self, lines):
        """Log a series of lines excluding trailing whitespace."""
        for line in lines:
            self.write(line)

    @property
    def is_debug_target(self):
        """Return True if this logger should receive debug messages."""
        return self._debug_target


class StdErrLog(Logger):
    """Class to log to standard error."""

    def __init__(self, filename, tag):
        Logger.__init__(self, filename, tag)
        self._debug_target = False

    def __call__(self, lines):
        """Log specified lines to stderr."""
        timestamp = time.strftime(Logger.TimeFormat,
                                  time.localtime(time.time()))
        if isinstance(lines, str):
            lines = [lines]

        for line in lines:
            print(f"{timestamp}{line}", file=sys.__stderr__)

        sys.__stderr__.flush()


class FileLog(Logger):
    """Class to log to a file."""

    def __init__(self, filename, tag):
        Logger.__init__(self, filename, tag)
        self._hostname = os.uname()[1]

    def __call__(self, lines):
        """Log specified lines to the file."""
        with open(self._logfile, "at", encoding="utf-8") as logf:
            timestamp = time.strftime(Logger.TimeFormat,
                                      time.localtime(time.time()))

            if isinstance(lines, str):
                lines = [lines]

            for line in lines:
                print(f"{timestamp}{self._hostname} {self._source}{line}", file=logf)


class LogFactory:
    """Singleton to log messages to various destinations."""

    log_methods = []
    have_stderr = False

    def add_file(self, filename, tag=None):
        """When logging messages, log them to specified file."""
        if filename:
            LogFactory.log_methods.append(FileLog(filename, tag))

    def add_stderr(self):
        """When logging messages, log them to standard error."""
        if not LogFactory.have_stderr:
            LogFactory.have_stderr = True
            LogFactory.log_methods.append(StdErrLog(None, None))

    def log(self, args):
        """Log a message (to all configured log destinations)."""
        for logfn in LogFactory.log_methods:
            logfn(args.strip())

    def debug(self, args):
        """Log a debug message (to all configured log destinations)."""
        for logfn in LogFactory.log_methods:
            if logfn.is_debug_target:
                logfn(f"debug: {args.strip()}")

    def traceback(self, traceback):
        """Log a stack trace (to all configured log destinations)."""
        for logfn in LogFactory.log_methods:
            traceback.print_exc(50, logfn)



"""Fail a random resource and verify its fail count increases."""

__copyright__ = "Copyright 2000-2025 the Pacemaker project contributors"
__license__ = "GNU General Public License version 2 or later (GPLv2+) WITHOUT ANY WARRANTY"

from pacemaker._cts.audits import AuditResource
from pacemaker._cts.tests.ctstest import CTSTest
from pacemaker._cts.tests.simulstartlite import SimulStartLite
from pacemaker._cts.tests.starttest import StartTest
from pacemaker._cts.timer import Timer

# Disable various pylint warnings that occur in so many places throughout this
# file it's easiest to just take care of them globally.  This does introduce the
# possibility that we'll miss some other cause of the same warning, but we'll
# just have to be careful.

# pylint doesn't understand that self._rsh is callable.
# pylint: disable=not-callable


class ResourceRecover(CTSTest):
    """Fail a random resource."""

    def __init__(self, cm):
        """
        Create a new ResourceRecover instance.

        Arguments:
        cm -- A ClusterManager instance
        """
        CTSTest.__init__(self, cm)

        self.benchmark = True
        self.name = "ResourceRecover"

        self._action = "asyncmon"
        self._interval = 0
        self._rid = None
        self._rid_alt = None
        self._start = StartTest(cm)
        self._startall = SimulStartLite(cm)

    def __call__(self, node):
        """Perform this test."""
        self.incr("calls")

        if not self._startall(None):
            return self.failure("Setup failed")

        # List all resources active on the node (skip test if none)
        resourcelist = self._cm.active_resources(node)
        if not resourcelist:
            self._logger.log(f"No active resources on {node}")
            return self.skipped()

        # Choose one resource at random
        rsc = self._choose_resource(node, resourcelist)
        if rsc is None:
            return self.failure(f"Could not get details of resource '{self._rid}'")

        if rsc.id == rsc.clone_id:
            self.debug(f"Failing {rsc.id}")
        else:
            self.debug(f"Failing {rsc.id} (also known as {rsc.clone_id})")

        # Log patterns to watch for (failure, plus restart if managed)
        pats = [
            self.templates["Pat:CloneOpFail"] % (self._action, rsc.id, rsc.clone_id)
        ]

        if rsc.managed:
            pats.append(self.templates["Pat:RscOpOK"] % ("stop", self._rid))

            if rsc.unique:
                pats.append(self.templates["Pat:RscOpOK"] % ("start", self._rid))
            else:
                # Anonymous clones may get restarted with a different clone number
                pats.append(self.templates["Pat:RscOpOK"] % ("start", ".*"))

        # Fail resource. (Ideally, we'd fail it twice, to ensure the fail count
        # is incrementing properly, but it might restart on a different node.
        # We'd have to temporarily ban it from all other nodes and ensure the
        # migration-threshold hasn't been reached.)
        if self._fail_resource(rsc, node, pats) is None:
            # self.failure() already called
            return None

        return self.success()

    def _choose_resource(self, node, resourcelist):
        """Choose a random resource to target."""
        self._rid = self._env.random_gen.choice(resourcelist)
        self._rid_alt = self._rid
        (_, lines) = self._rsh(node, "crm_resource -c", verbose=1)

        for line in lines:
            if line.startswith("Resource: "):
                rsc = AuditResource(self._cm, line)

                if rsc.id == self._rid:
                    # Handle anonymous clones that get renamed
                    self._rid = rsc.clone_id
                    return rsc

        return None

    def _get_failcount(self, node):
        """Check the fail count of targeted resource on given node."""
        cmd = "crm_failcount --quiet --query --resource %s --operation %s --interval %d --node %s"
        (rc, lines) = self._rsh(node, cmd % (self._rid, self._action, self._interval, node),
                                verbose=1)

        if rc != 0 or len(lines) != 1:
            lines = [line.strip() for line in lines]
            s = " // ".join(lines)
            self._logger.log(f"crm_failcount on {node} failed ({rc}): {s}")
            return -1

        try:
            failcount = int(lines[0])
        except (IndexError, ValueError):
            s = " ".join(lines)
            self._logger.log(f"crm_failcount output on {node} unparseable: {s}")
            return -1

        return failcount

    def _fail_resource(self, rsc, node, pats):
        """Fail the targeted resource, and verify as expected."""
        orig_failcount = self._get_failcount(node)

        watch = self.create_watch(pats, 60)
        watch.set_watch()

        self._rsh(node, f"crm_resource -V -F -r {self._rid} -H {node} &>/dev/null")

        with Timer(self._logger, self.name, "recover"):
            watch.look_for_all()

        self._cm.cluster_stable()
        recovered = self._cm.resource_location(self._rid)

        if watch.unmatched:
            return self.failure(f"Patterns not found: {watch.unmatched!r}")

        if rsc.unique and len(recovered) > 1:
            return self.failure(f"{self._rid} is now active on more than one node: {recovered!r}")

        if recovered:
            self.debug(f"{self._rid} is running on: {recovered!r}")

        elif rsc.managed:
            return self.failure(f"{self._rid} was not recovered and is inactive")

        new_failcount = self._get_failcount(node)
        if new_failcount != orig_failcount + 1:
            return self.failure(f"{self._rid} fail count is {new_failcount} not {orig_failcount + 1}")

        # Anything but None is success
        return 0

    @property
    def errors_to_ignore(self):
        """Return a list of errors which should be ignored."""
        return [
            f"Updating failcount for {self._rid}",
            fr"schedulerd.*: Recover\s+({self._rid}|{self._rid_alt})\s+\(.*\)",
            r"Unknown operation: fail",
            self.templates["Pat:RscOpOK"] % (self._action, self._rid),
            f"(ERROR|error).*: Action {self._rid}_{self._action}_{self._interval} .* initiated outside of a transition",
        ]



"""A module providing constants relating to why a process or function exited."""

__all__ = ["ExitStatus"]
__copyright__ = "Copyright 2023-2025 the Pacemaker project contributors"
__license__ = "GNU Lesser General Public License version 2.1 or later (LGPLv2.1+)"

from enum import IntEnum, unique

from pacemaker._library import _libcrmcommon


# These values must be kept in sync with include/crm/common/results.h
@unique
class ExitStatus(IntEnum):
    """
    Exit status codes for a function or process.

    These constants describe both success and failure conditions.
    """

    OK = 0
    ERROR = 1
    INVALID_PARAM = 2
    UNIMPLEMENT_FEATURE = 3
    INSUFFICIENT_PRIV = 4
    NOT_INSTALLED = 5
    NOT_CONFIGURED = 6
    NOT_RUNNING = 7
    PROMOTED = 8
    FAILED_PROMOTED = 9
    USAGE = 64
    DATAERR = 65
    NOINPUT = 66
    NOUSER = 67
    NOHOST = 68
    UNAVAILABLE = 69
    SOFTWARE = 70
    OSERR = 71
    OSFILE = 72
    CANTCREAT = 73
    IOERR = 74
    TEMPFAIL = 75
    PROTOCOL = 76
    NOPERM = 77
    CONFIG = 78
    FATAL = 100
    PANIC = 101
    DISCONNECT = 102
    OLD = 103
    DIGEST = 104
    NOSUCH = 105
    QUORUM = 106
    UNSAFE = 107
    EXISTS = 108
    MULTIPLE = 109
    EXPIRED = 110
    NOT_YET_IN_EFFECT = 111
    INDETERMINATE = 112
    UNSATISFIED = 113
    NO_DC = 114
    TIMEOUT = 124
    DEGRADED = 190
    DEGRADED_PROMOTED = 191
    NONE = 193
    MAX = 255

    def __str__(self):
        """Given an ExitStatus, return the matching error string."""
        return _libcrmcommon.crm_exit_str(self.value).decode()



"""Pattern-holding classes for Pacemaker's Cluster Test Suite (CTS)."""

__all__ = ["PatternSelector"]
__copyright__ = "Copyright 2008-2025 the Pacemaker project contributors"
__license__ = "GNU General Public License version 2 or later (GPLv2+)"

import argparse

from pacemaker.buildoptions import BuildOptions


class BasePatterns:
    """
    The base class for holding a stack-specific set of command and log file/stdout patterns.

    Stack-specific classes need to be built on top of this one.
    """

    def __init__(self):
        """Create a new BasePatterns instance which holds a very minimal set of basic patterns."""
        self._bad_news = []
        self._components = {}
        self._name = "crm-base"

        self._ignore = [
            "avoid confusing Valgrind",

            # Logging bug in some versions of libvirtd
            r"libvirtd.*: internal error: Failed to parse PCI config address",

            # pcs can log this when node is fenced, but fencing is OK in some
            # tests (and we will catch it in pacemaker logs when not OK)
            r"pcs.daemon:No response from: .* request: get_configs, error:",

            # This is overbroad, but there's no way to say that only certain
            # transition errors are acceptable. We have to rely on causes of a
            # transition error logging their own error message, which should
            # always be the case.
            r"pacemaker-schedulerd.* Calculated transition .*/pe-error",

            # This message comes up periodically but doesn't actually seem to
            # be related to any specific test failure, so just ignore it.
            r"pacemaker-based.* Local CIB .* differs from",
        ]

        self._commands = {
            "StatusCmd": "crmadmin -t 60 -S %s 2>/dev/null",
            "CibQuery": "cibadmin -Q",
            "CibAddXml": "cibadmin --modify -c --xml-text %s",
            "CibDelXpath": "cibadmin --delete --xpath %s",
            "RscRunning": BuildOptions.DAEMON_DIR + "/cts-exec-helper -R -r %s",
            "CIBfile": "%s:" + BuildOptions.CIB_DIR + "/cib.xml",
            "TmpDir": "/tmp",

            "BreakCommCmd": "iptables -A INPUT -s %s -j DROP >/dev/null 2>&1",
            "FixCommCmd": "iptables -D INPUT -s %s -j DROP >/dev/null 2>&1",

            "MaintenanceModeOn": "cibadmin --modify -c --xml-text '<cluster_property_set id=\"cib-bootstrap-options\"><nvpair id=\"cts-maintenance-mode-setting\" name=\"maintenance-mode\" value=\"true\"/></cluster_property_set>'",
            "MaintenanceModeOff": "cibadmin --delete --xpath \"//nvpair[@name='maintenance-mode']\"",

            "StandbyCmd": "crm_attribute -Vq  -U %s -n standby -l forever -v %s 2>/dev/null",
            "StandbyQueryCmd": "crm_attribute -qG -U %s -n standby -l forever -d off 2>/dev/null",
        }

        self._search = {
            "Pat:DC_IDLE": r"pacemaker-controld.*State transition.*-> S_IDLE",

            # This won't work if we have multiple partitions
            "Pat:Local_started": r"%s\W.*controller successfully started",
            "Pat:NonDC_started": r"%s\W.*State transition.*-> S_NOT_DC",
            "Pat:DC_started": r"%s\W.*State transition.*-> S_IDLE",
            "Pat:We_stopped": r"%s\W.*OVERRIDE THIS PATTERN",
            "Pat:They_stopped": r"%s\W.*LOST:.* %s ",
            "Pat:They_dead": r"node %s.*: is dead",
            "Pat:They_up": r"%s %s\W.*OVERRIDE THIS PATTERN",
            "Pat:TransitionComplete": "Transition status: Complete: complete",

            "Pat:Fencing_start": r"Requesting peer fencing .* targeting %s",
            "Pat:Fencing_ok": r"pacemaker-fenced.*:\s*Operation .* targeting %s by .* for .*@.*: OK",
            "Pat:Fencing_recover": r"pacemaker-schedulerd.*: Recover\s+%s",
            "Pat:Fencing_active": r"stonith resource .* is active on 2 nodes (attempting recovery)",
            "Pat:Fencing_probe": r"pacemaker-controld.* Result of probe operation for %s on .*: Error",

            "Pat:RscOpOK": r"pacemaker-controld.*:\s+Result of %s operation for %s.*: (0 \()?OK",
            "Pat:RscOpFail": r"pacemaker-schedulerd.*:.*Unexpected result .* recorded for %s of %s ",
            "Pat:CloneOpFail": r"pacemaker-schedulerd.*:.*Unexpected result .* recorded for %s of (%s|%s) ",
            "Pat:RscRemoteOpOK": r"pacemaker-controld.*:\s+Result of %s operation for %s on %s: (0 \()?OK",
            "Pat:NodeFenced": r"pacemaker-controld.*:\s* Peer %s was terminated \(.*\) by .* on behalf of .*: OK",
        }

    def get_component(self, key):
        """
        Return the patterns for a single component as a list, given by key.

         This is typically the name of some subprogram (pacemaker-based,
         pacemaker-fenced, etc.) or various special purpose keys.  If key is
         unknown, return an empty list.
        """
        if key in self._components:
            return self._components[key]

        print(f"Unknown component '{key}' for {self._name}")
        return []

    def get_patterns(self, key):
        """
        Return various patterns supported by this object, given by key.

        Depending on the key, this could either be a list or a hash.  If key is
        unknown, return None.
        """
        if key == "BadNews":
            return self._bad_news
        if key == "BadNewsIgnore":
            return self._ignore
        if key == "Commands":
            return self._commands
        if key == "Search":
            return self._search
        if key == "Components":
            return self._components

        print(f"Unknown pattern '{key}' for {self._name}")
        return None

    def __getitem__(self, key):
        if key == "Name":
            return self._name
        if key in self._commands:
            return self._commands[key]
        if key in self._search:
            return self._search[key]

        print(f"Unknown template '{key}' for {self._name}")
        return None


class Corosync2Patterns(BasePatterns):
    """Patterns for Corosync version 2 cluster manager class."""

    def __init__(self):
        BasePatterns.__init__(self)
        self._name = "crm-corosync"

        self._commands.update({
            "StartCmd": "service corosync start && service pacemaker start",
            "StopCmd": "service pacemaker stop; [ ! -e /usr/sbin/pacemaker-remoted ] || service pacemaker_remote stop; service corosync stop",

            "EpochCmd": "crm_node -e",
            "QuorumCmd": "crm_node -q",
            "PartitionCmd": "crm_node -p",
        })

        self._search.update({
            # Close enough ... "Corosync Cluster Engine exiting normally" isn't
            # printed reliably.
            "Pat:We_stopped": r"%s\W.*Unloading all Corosync service engines",
            "Pat:They_stopped": r"%s\W.*pacemaker-controld.*Node %s(\[|\s).*state is now lost",
            "Pat:They_dead": r"pacemaker-controld.*Node %s(\[|\s).*state is now lost",
            "Pat:They_up": r"\W%s\W.*pacemaker-controld.*Node %s state is now member",

            "Pat:ChildExit": r"\[[0-9]+\] exited with status [0-9]+ \(",
            # "with signal 9" == pcmk_child_exit(), "$" == check_active_before_startup_processes()
            "Pat:ChildKilled": r"%s\W.*pacemakerd.*%s\[[0-9]+\] terminated( with signal 9|$)",
            "Pat:ChildRespawn": r"%s\W.*pacemakerd.*Respawning subdaemon %s after unexpected exit",

            "Pat:InfraUp": r"%s\W.*corosync.*Initializing transport",
            "Pat:PacemakerUp": r"%s\W.*pacemakerd.*Starting Pacemaker",
        })

        self._ignore += [
            r"crm_mon:",
            r"crmadmin:",
            r"update_trace_data",
            r"async_notify:.*strange, client not found",
            r"Parse error: Ignoring unknown option .*nodename",
            r"error.*: Operation 'reboot' .* using FencingFail returned ",
            r"getinfo response error: 1$",
            r"sbd.* error: inquisitor_child: DEBUG MODE IS ACTIVE",
            r"sbd.* pcmk:\s*error:.*Connection to cib_ro.* (failed|closed)",
        ]

        self._bad_news = [
            r"[^(]error:",
            r"crit:",
            r"ERROR:",
            r"CRIT:",
            r"Shutting down...NOW",
            r"Timer I_TERMINATE just popped",
            r"input=I_ERROR",
            r"input=I_FAIL",
            r"input=I_INTEGRATED cause=C_TIMER_POPPED",
            r"input=I_FINALIZED cause=C_TIMER_POPPED",
            r"input=I_ERROR",
            r"(pacemakerd|pacemaker-execd|pacemaker-controld):.*, exiting",
            r"schedulerd.*Attempting recovery of resource",
            r"is taking more than 2x its timeout",
            r"Confirm not received from",
            r"Welcome reply not received from",
            r"Attempting to schedule .* after a stop",
            r"Resource .* was active at shutdown",
            r"duplicate entries for call_id",
            r"Search terminated:",
            r":global_timer_callback",
            r"Faking parameter digest creation",
            r"Parameters to .* action changed:",
            r"Parameters to .* changed",
            r"pacemakerd.*\[[0-9]+\] terminated( with signal|$)",
            r"pacemakerd.*\[[0-9]+\] .* will now be killed",
            r"pacemaker-schedulerd.*Recover\s+.*\(.* -\> .*\)",
            r"rsyslogd.* lost .* due to rate-limiting",
            r"Peer is not part of our cluster",
            r"We appear to be in an election loop",
            r"Unknown node -> we will not deliver message",
            r"(Blackbox dump requested|Problem detected)",
            r"pacemakerd.*Could not connect to Cluster Configuration Database API",
            r"Receiving messages from a node we think is dead",
            r"share the same cluster nodeid",
            r"share the same name",

            r"pacemaker-controld:.*Transition failed: terminated",
            r"Local CIB .* differs from .*:",
            r"warn.*:\s*Continuing but .* will NOT be used",
            r"warn.*:\s*Cluster configuration file .* is corrupt",
            r"Election storm",
            r"stalled the FSA with pending inputs",
        ]

        self._components["common-ignore"] = [
            r"Pending action:",
            r"resource( was|s were) active at shutdown",
            r"pending LRM operations at shutdown",
            r"Lost connection to the CIB manager",
            r"pacemaker-controld.*:\s*Action A_RECOVER .* not supported",
            r"pacemaker-controld.*:\s*Exiting now due to errors",
            r".*:\s*Requesting fencing \([^)]+\) targeting node ",
            r"(Blackbox dump requested|Problem detected)",
        ]

        self._components["corosync-ignore"] = [
            r"Could not connect to Corosync CFG: CS_ERR_LIBRARY",
            r"error:.*Connection to the CPG API failed: Library error",
            r"\[[0-9]+\] exited with status [0-9]+ \(",
            r"\[[0-9]+\] terminated with signal 15",
            r"pacemaker-based.*error:.*Corosync connection lost",
            r"pacemaker-fenced.*error:.*Corosync connection terminated",
            r"pacemaker-controld.*State transition .* S_RECOVERY",
            r"pacemaker-controld.*error:.*Input (I_ERROR|I_TERMINATE ) .*received in state",
            r"pacemaker-controld.*error:.*Could not recover from internal error",
            r"error:.*Connection to cib_(shm|rw).* (failed|closed)",
            r"error:.*cib_(shm|rw) IPC provider disconnected while waiting",
            r"error:.*Connection to (fencer|stonith-ng).* (closed|failed|lost)",
            r"error: Lost fencer connection",
        ]

        self._components["corosync"] = [
            # We expect each daemon to lose its cluster connection.
            # However, if the CIB manager loses its connection first,
            # it's possible for another daemon to lose that connection and
            # exit before losing the cluster connection.
            r"pacemakerd.*:\s*warning:.*Lost connection to cluster layer",
            r"pacemaker-attrd.*:\s*(crit|error):.*Lost connection to (Corosync process group|the CIB manager)",
            r"pacemaker-based.*:\s*crit:.*Exiting immediately after losing connection to cluster layer",
            r"pacemaker-controld.*:\s*(crit|error):.*Lost connection to (cluster layer|the CIB manager)",
            r"pacemaker-fenced.*:\s*(crit|error):.*Lost connection to (cluster layer|the CIB manager)",
            r"schedulerd.*Scheduling node .* for fencing",
            r"pacemaker-controld.*:\s*Peer .* was terminated \(.*\) by .* on behalf of .*:\s*OK",
        ]

        self._components["pacemaker-based"] = [
            r"pacemakerd.* pacemaker-attrd\[[0-9]+\] exited with status 102",
            r"pacemakerd.* pacemaker-controld\[[0-9]+\] exited with status 1",
            r"pacemakerd.* Respawning subdaemon pacemaker-attrd after unexpected exit",
            r"pacemakerd.* Respawning subdaemon pacemaker-based after unexpected exit",
            r"pacemakerd.* Respawning subdaemon pacemaker-controld after unexpected exit",
            r"pacemakerd.* Respawning subdaemon pacemaker-fenced after unexpected exit",
            r"pacemaker-.* Connection to cib_.* (failed|closed)",
            r"pacemaker-attrd.*:.*Lost connection to the CIB manager",
            r"pacemaker-controld.*:.*Lost connection to the CIB manager",
            r"pacemaker-controld.*I_ERROR.*handle_cib_disconnect",
            r"pacemaker-controld.* State transition .* S_RECOVERY",
            r"pacemaker-controld.*: Input I_TERMINATE .*from do_recover",
            r"pacemaker-controld.*Could not recover from internal error",
        ]

        self._components["pacemaker-based-ignore"] = [
            r"pacemaker-execd.*Connection to (fencer|stonith-ng).* (closed|failed|lost)",
            r"pacemaker-controld.*:\s+Result of .* operation for Fencing.*Error \(Lost connection to fencer\)",
            r"pacemaker-controld.*:Could not connect to attrd: Connection refused",
        ]

        self._components["pacemaker-execd"] = [
            r"pacemaker-controld.*Lost connection to local executor",
            r"pacemaker-controld.*I_ERROR.*lrm_connection_destroy",
            r"pacemaker-controld.*State transition .* S_RECOVERY",
            r"pacemaker-controld.*: Input I_TERMINATE .*from do_recover",
            r"pacemaker-controld.*Could not recover from internal error",
            r"pacemakerd.*pacemaker-controld\[[0-9]+\] exited with status 1",
            r"pacemakerd.* Respawning subdaemon pacemaker-execd after unexpected exit",
            r"pacemakerd.* Respawning subdaemon pacemaker-controld after unexpected exit",
        ]

        self._components["pacemaker-execd-ignore"] = [
            r"pacemaker-(attrd|controld).*Connection to lrmd.* (failed|closed)",
            r"pacemaker-(attrd|controld).*Could not execute alert",
        ]

        self._components["pacemaker-controld"] = [
            r"State transition .* -> S_IDLE",
        ]

        self._components["pacemaker-controld-ignore"] = []
        self._components["pacemaker-attrd"] = []
        self._components["pacemaker-attrd-ignore"] = [
            r"pacemaker-controld.*Connection to attrd (IPC failed|closed)",
        ]

        self._components["pacemaker-schedulerd"] = [
            r"State transition .* S_RECOVERY",
            r"pacemakerd.* Respawning subdaemon pacemaker-controld after unexpected exit",
            r"pacemaker-controld\[[0-9]+\] exited with status 1 \(",
            r"pacemaker-controld.*Lost connection to the scheduler",
            r"pacemaker-controld.*I_ERROR.*save_cib_contents",
            r"pacemaker-controld.*: Input I_TERMINATE .*from do_recover",
            r"pacemaker-controld.*Could not recover from internal error",
        ]

        self._components["pacemaker-schedulerd-ignore"] = [
            r"Connection to pengine.* (failed|closed)",
        ]

        self._components["pacemaker-fenced"] = [
            r"error:.*Connection to (fencer|stonith-ng).* (closed|failed|lost)",
            r"Lost fencer connection",
            r"pacemaker-controld.*Fencer successfully connected",
        ]

        self._components["pacemaker-fenced-ignore"] = [
            r"(error|warning):.*Connection to (fencer|stonith-ng).* (closed|failed|lost)",
            r"error:.*Lost fencer connection",
            r"error:.*Fencer connection failed \(will retry\)",
            r"pacemaker-controld.*:\s+Result of .* operation for Fencing.*Error \(Lost connection to fencer\)",
        ]

        self._components["pacemaker-fenced-ignore"].extend(self._components["common-ignore"])


patternVariants = {
    "crm-base": BasePatterns,
    "crm-corosync": Corosync2Patterns
}


class PatternSelector:
    """Choose from among several Pattern objects and return the information from that object."""

    def __init__(self, name="crm-corosync"):
        """
        Create a new PatternSelector object.

        Instantiate whatever class is given by name.  Defaults to Corosync2Patterns
        for "crm-corosync" or None.  While other objects could be supported in the
        future, only this and the base object are supported at this time.
        """
        self._name = name

        # If no name was given, use the default.  Otherwise, look up the appropriate
        # class in patternVariants, instantiate it, and use that.
        if not name:
            self._base = Corosync2Patterns()
        else:
            self._base = patternVariants[name]()

    def get_patterns(self, kind):
        """Call get_patterns on the previously instantiated pattern object."""
        return self._base.get_patterns(kind)

    def get_template(self, key):
        """
        Return a single pattern from the previously instantiated pattern object.

        If no pattern exists for the given key, return None.
        """
        return self._base[key]

    def get_component(self, kind):
        """Call get_component on the previously instantiated pattern object."""
        return self._base.get_component(kind)

    def __getitem__(self, key):
        """Return the pattern for the given key, or None if it does not exist."""
        return self.get_template(key)


# PYTHONPATH=python python python/pacemaker/_cts/patterns.py -k crm-corosync -t StartCmd
if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument("-k", "--kind", metavar="KIND")
    parser.add_argument("-t", "--template", metavar="TEMPLATE")

    args = parser.parse_args()

    print(PatternSelector(args.kind)[args.template])
